/* OTA-ioLykta-wifi
*The MIT License (MIT)
*
*Copyright (c) 2017 ruhigundrelaxed (Martin Hilgers)

*Permission is hereby granted, free of charge, 
*to any person obtaining a copy of this software and associated documentation files (the "Software"), 
*to deal in the Software without restriction, 
*including without limitation the rights to 
*use, copy, modify, merge, publish, distribute  of the Software, 
*and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
*The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
*THE SOFTWARE IS PROVIDED "AS IS", 
*WITHOUT WARRANTY OF ANY KIND, 
*EXPRESS OR IMPLIED, 
*INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
*FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
*DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
*ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*
*The Software was inspired by the great work of: 
*Hans Luijten (t4a_boblight) and 
*Pascal Gollor (OTA-mDNS-SPIFFS) .
*/



// includes
#include <ESP8266WiFi.h>          // The ESPWifi lib
#include <ESP8266mDNS.h>          // THE ESPmDNS lib used for OTA
#include <WiFiUdp.h>              // Wifi UDP lib used for OTA
#include <FS.h>                   // The FS lib for spiffs
#include <ArduinoOTA.h>           // The Arduino OTA lib
#include <Adafruit_NeoPixel.h>    // The Neopixel lib
//#include <DallasTemperature.h>  // For Temperature sensing

//*******************************************************************************************************
//*************************************Config************************************************************
//*******************************************************************************************************
//Networking Stuff
#define HOSTNAME "ioLykta-"       // Hostename prefix. The setup function adds the Chip ID at the end.
#define MAX_SRV_CLIENTS 1         // Only one Client should be able to connect to the server
#define BOBPORT 4711              // The TCP Port 
//run the following line on the machine running the boblightd
//sudo socat -d -d pty,link=/dev/ttyREMOTE,raw,echo=0 tcp:192.168.123.27:4711
//when ttyREMOTE is in your boblight.conf
//
//Pin assignment
#define WS2812PIN 14              // The Pin the WS2812 is connected to
#define POWERSWITCHPIN 13         // The Pin the Mosfet is connected to
#define ONEWIREPIN 12             // The Pin the DS18B20 is connected to

//Boblight Configuration Stuff
#define STARTCOLOR 0x222222       // LED colors at start
#define BLACK      0x000000       // LED color BLACK
#define LEDCOUNT   144            // Number of LEDs used for boblight
#define SHOWDELAY  200            // Delay in micro seconds before showing
#define BRIGHTNESS 90             // Max. brightness in %

//Serial Debugging Stuff
#define DEBUG
#define BAUDRATE   9600           // Serial port speed
//*******************************************************************************************************
//*******************************************************************************************************
//*******************************************************************************************************


//States
#define STATE_WAITING   1         // Waiting for prefix
#define STATE_DO_PREFIX 2         // Processing prefix
#define STATE_DO_DATA   3         // Receiving LED colors



//Let us init the strip
Adafruit_NeoPixel strip = Adafruit_NeoPixel(LEDCOUNT, WS2812PIN, NEO_GRB + NEO_KHZ800);

//Init the Socket
WiFiServer server(4711);
WiFiClient serverClients[MAX_SRV_CLIENTS];

//Init the Temperature Sensor
//OneWire oneWire(ONEWIREPIN);
//DallasTemperature sensors(&oneWire); 


int state;                      // Init the state holding var
int readData;                   // Read the data (1)
int currentLED;                 // Needed for assigning the color to the right LED


const char* ap_default_ssid = "streng";       // Default SSID.
const char* ap_default_psk = "geheim"; // Default PSK.
const char prefix[] = {0x41, 0x64, 0x61, 0x00, 0x18, 0x4D}; 

char buffer[sizeof(prefix)];



void setup()
{
  
  #ifdef DEBUG
    Serial.begin(BAUDRATE);                         // Init serial speed    
    Serial.println("INFO: Program start! Setting pinmode 4 Powerswitch and switch Power on!");
  #endif

  pinMode(POWERSWITCHPIN, OUTPUT);                  //Set Pinmode for the FET
  digitalWrite(POWERSWITCHPIN, HIGH);               //Switch on FET


 //sensors.begin();                                   //Init the Temp Sensor(s)
 server.begin();
  #ifdef DEBUG
    Serial.println("INFO: Init Strip and set Startcolor! Init the state var!");
  #endif
  strip.begin();                                    //Init strip.
  strip.setBrightness( (255 / 100) * BRIGHTNESS );  //Set LED Brightness
  setAllLEDs(BLACK, 0);                             //Set all Leds to BLACK  
  setAllLEDs(STARTCOLOR, 5);                        //Set all Leds to Startcolor
  state = STATE_WAITING;                            // Initial state: Waiting for prefix

  #ifdef DEBUG
    Serial.println("INFO: Setting the Hostname!");
  #endif  
  String station_ssid = "";
  String station_psk = "";
  String hostname(HOSTNAME); 
  hostname += String(ESP.getChipId(), HEX);
  WiFi.hostname(hostname);



  #ifdef DEBUG
    Serial.println("INFO: Open Filesystem for config reading!");
  #endif  

  // Initialize file system.
  if (!SPIFFS.begin())
  {
    #ifdef DEBUG
        Serial.println("ERROR: Could not mount fs!");
    #endif  
    return;
  }
  #ifdef DEBUG
    Serial.println("INFO: Load wifi connection information.!");
  #endif  
  if (!loadConfig(&station_ssid, &station_psk))
  {
    #ifdef DEBUG
      Serial.println("ERROR: Could not read config. Init vars with NULL");
    #endif  
    station_ssid = "";
    station_psk = "";
  }

  // Check WiFi connection
  // ... check mode
  if (WiFi.getMode() != WIFI_STA)
  {
    WiFi.mode(WIFI_STA);
    delay(10);
  }

  // ... Compare file config with sdk config.
  if (WiFi.SSID() != station_ssid || WiFi.psk() != station_psk)
  {
    Serial.println("INFO: WiFi config changed.");

    // ... Try to connect to WiFi station.
    WiFi.begin(station_ssid.c_str(), station_psk.c_str());
  }
  else
  {
    // ... Begin with sdk config.
    WiFi.begin();
  }

  // ... Give ESP 10 seconds to connect to station.
  unsigned long startTime = millis();
  while (WiFi.status() != WL_CONNECTED && millis() - startTime < 10000)
  {
    delay(500);
  }
  
  // Check connection
  if(WiFi.status() == WL_CONNECTED)
  {
    #ifdef DEBUG
      Serial.println("INFO: GOT IP:");
      Serial.println(WiFi.localIP());
    #endif     
  }
  else
  {
    //Serial.println("Can not connect to WiFi station. Go into AP mode.");
     #ifdef DEBUG
      Serial.println("INFO: Can not connect to WiFi station. Go into AP mode.");
     #endif  
    WiFi.mode(WIFI_AP);
    delay(10);
    WiFi.softAP(ap_default_ssid, ap_default_psk);
    #ifdef DEBUG
      Serial.println("INFO: GOT IP:");
      Serial.println(WiFi.softAPIP());
    #endif 
  }
#ifdef DEBUG
  Serial.println("INFO: Start the OTA server and finish the Setup routine. Entering Main Loop...");
#endif  
  ArduinoOTA.setHostname((const char *)hostname.c_str());
  ArduinoOTA.begin();
}


/**
 * @brief Arduino loop function.
 */
void loop()
{
  // Handle OTA server.
  ArduinoOTA.handle();
  yield();
  //int clientcount;
  
  //Start the TCP Socketroutine
  
  
  if (server.hasClient()){
  //  for(clientcount = 0; clientcount < MAX_SRV_CLIENTS; clientcount++){
       if (!serverClients[0] || !serverClients[0].connected()){
        if(serverClients[0]) serverClients[0].stop();
        serverClients[0] = server.available();
        #ifdef DEBUG
          Serial.print("INFO: New client: "); Serial.print(0);
        #endif
        //continue; 
       }
      }
    WiFiClient serverClient = server.available();
    serverClient.stop();
   // }

    //for(clientcount = 0; clientcount < MAX_SRV_CLIENTS; clientcount++){
    if (serverClients[0] && serverClients[0].connected()){
      if(serverClients[0].available()){
        
        switch(state)
        {
            case STATE_WAITING:                                        // *** Waiting for prefix ***
                if( serverClients[0].available()>0 )
                {
                  readData = serverClients[0].read();       // Read one character
                  if ( readData == prefix[0] )                        // if this character is 1st prefix char
                     { state = STATE_DO_PREFIX; }                     // then set state to handle prefix
                }
                break;

            case STATE_DO_PREFIX:                                     // *** Processing Prefix ***
                if( serverClients[0].available() > sizeof(prefix) - 2 ) 
                {
                   serverClients[0].readBytes(buffer, sizeof(prefix) - 1);
                   for( int Counter = 0; Counter < sizeof(prefix) - 1; Counter++) 
                   {
                      if( buffer[Counter] == prefix[Counter+1] ) 
                      {
                        state = STATE_DO_DATA;                        // Received character is in prefix, continue
                        currentLED = 0;                               // Set current LED to the first one
                      }
                      else 
                      {
                        state = STATE_WAITING;                        // Crap, one of the received chars is NOT in the prefix
                        break;                                        // Exit, to go back to waiting for the prefix
                      }                                               // end if buffer
                  }                                                   // end for Counter
                }                                                     // end if ClientsData
                break;

            case STATE_DO_DATA:                                             // *** Process incoming color data ***
                if( serverClients[0].available() > 2 )            // if we receive more than 2 chars
                {
                  serverClients[0].readBytes( buffer, 3 );        // Abuse buffer to temp store 3 charaters
                  strip.setPixelColor( currentLED++, buffer[0], buffer[1], buffer[2]);  // and assing to LEDs
                }

                if( currentLED > LEDCOUNT )        // Reached the last LED? Display it!
                {
                   strip.show();                  // Make colors visible
                   delayMicroseconds(SHOWDELAY);  // Wait a few micro seconds
                   state = STATE_WAITING;         // Reset to waiting ...
                   currentLED = 0;                // and go to LED one
                   break;                         // and exit ... and do it all over again
                }
                break; 
          } // switch(state)
    } //endif Data avail
  }   //endif clients connected
// }    //for each client     
}  

 


bool saveConfig(String *ssid, String *pass)
{
  // Open config file for writing.
  File configFile = SPIFFS.open("/cl_conf.txt", "w");
  if (!configFile)
  {
    //Serial.println("Failed to open cl_conf.txt for writing");

    return false;
  }

  // Save SSID and PSK.
  configFile.println(*ssid);
  configFile.println(*pass);

  configFile.close();
  
  return true;
} // saveConfig



bool loadConfig(String *ssid, String *pass)
{
  // open file for reading.
  File configFile = SPIFFS.open("/cl_conf.txt", "r");
  if (!configFile)
  {
    //Serial.println("Failed to open cl_conf.txt.");

    return false;
  }

  // Read content from config file.
  String content = configFile.readString();
  configFile.close();
  
  content.trim();

  // Check if ther is a second line available.
  int8_t pos = content.indexOf("\r\n");
  uint8_t le = 2;
  // check for linux and mac line ending.
  if (pos == -1)
  {
    le = 1;
    pos = content.indexOf("\n");
    if (pos == -1)
    {
      pos = content.indexOf("\r");
    }
  }

  // If there is no second line: Some information is missing.
  if (pos == -1)
  {
    //Serial.println("Invalid content.");
    //Serial.println(content);

    return false;
  }

  // Store SSID and PSK into string vars.
  *ssid = content.substring(0, pos);
  *pass = content.substring(pos + le);

  ssid->trim();
  pass->trim();


  return true;
} // loadConfig


void setAllLEDs(uint32_t color, int wait)
{
  for ( int Counter=0; Counter < LEDCOUNT; Counter++ )      // For each LED
  {
    strip.setPixelColor( Counter, color );                  // .. set the color

    if( wait > 0 )                                          // if a wait time was set then
    {
      strip.show();                                         // Show the LED color
      delay(wait);                                          // and wait before we do the next LED
    } // if wait

  } // for Counter

  strip.show();                                             // Show all LEDs
} // setAllLEDs



